# abi-check.py
import csv
import os
import sys
import datetime
import dnf
import dnf.base
import dnf.conf
import dnf.cli.term
from dnf.cli import CliError
from dnf.cli.output import Output
from dnfpluginscore import _, logger

sys.path.append('/usr/share/abi-check')
from utils.output import ABIOutput
from utils import translate
from utils import utils


class ABICheck(dnf.Plugin):
    """
    实现升级或降级时的兼容性检查
    """
    name = 'abi-check'

    def __init__(self, base, cli):
        super(ABICheck, self).__init__(base, cli)
        self.base = base
        self.cli = cli
        self.is_compatible = True  # 整体结果是否兼容
        self.undo = True  # 整体结果是否能够回退
        self.conf = None
        self.package_info = []
        self.is_undo_flag = False  # 记录是否是回退操作
        self.details_file = None
        self.base_dir = "/var/lib/dnf/abi-check"  # 存放详细结果的默认路径

    @staticmethod
    def _get_output(package_info):
        """
        得到格式化输出的字符串并打印到终端
        :param package_info: 存放PackageInfo实例的列表
        """
        output_obj = ABIOutput(package_info)
        output_str = output_obj.get_output()
        logger.info(output_str)

    def _get_compatibility_details(self, change, detail):
        """
        获取当前安装版本、将安装版本、变化详情
        :param change: 两个版本包的差异文件
        :param detail: 差异的详情
        """
        if isinstance(detail, dict):
            if self.is_undo_flag:  # 如果是回退，则需要将so的版本调个头
                installed = change.split(":")[1]
                install = change.split(":")[0]
                detail_msg = detail["detail"]
            else:
                installed = change.split(":")[0]
                install = change.split(":")[1]
                detail_msg = detail["detail"]
        else:  # 表示是内核软件包的kabi变化信息
            installed = []
            install = []
            detail_msg = ""
            if self.is_undo_flag:  # 如果是回退，则需要将kabi的变化调个头
                for line in detail:
                    installed.append(line.split(":")[1])
                    install.append(line.split(":")[0])
            else:
                for line in detail:
                    installed.append(line.split(":")[0])
                    install.append(line.split(":")[1])
            installed = '\n'.join(installed)
            install = '\n'.join(install)

        # 对无数据的字段填充值N/A
        installed = installed if installed else 'N/A'
        install = install if install else 'N/A'
        detail_msg = detail_msg if detail_msg else 'N/A'
        return installed, install, detail_msg

    def _write_check_result(self):
        """
        将更详细的检查结果写入到文件中
        """

        # 将详细信息写入到文件中
        self.details_file = "compatibility-details-{date}.csv".format(
            date=datetime.datetime.strftime(datetime.datetime.now(), "%Y%m%d-%H%M%S"))

        if not os.path.exists(self.base_dir):
            os.makedirs(self.base_dir)

        self.details_file = os.path.join(self.base_dir, self.details_file)
        with open(self.details_file, "w", newline='', encoding='utf-8') as fd:
            csvwriter = csv.writer(fd)
            # 写入表头
            errata_msg = "Upgrade or downgrade related errata"
            cve_msg = "CVEs or bugs involved in upgrade or downgrade"
            head = [translate("Package"), translate(errata_msg), translate("URL"), translate("Severity"),
                    translate(cve_msg), translate("Compatibility for ABI or KABI"),
                    translate("Compatibility for Drive KABI"), translate("Compatible"), translate("Undo"),
                    translate("Installed package"), translate("Install package"),
                    translate("Installed version of ABI or KABI"), translate("Install version of ABI or KABI"),
                    translate("Compatibility Details")]
            write_content = [head]
            for package in self.package_info:
                tmp_content = []
                compatibility_abi = package.compatibility_abi
                compatibility_drive_kabi = package.compatibility_drive_kabi if package.compatibility_drive_kabi else 'N/A'
                is_compatible = translate("yes") if package.is_compatible else translate('no')
                undo = translate("yes") if package.undo else translate("no")
                for severity, data_list in package.erratas.items():
                    for data in data_list:
                        errata = list(data.keys())[0]
                        cve_and_bug = data[errata]['CVE']
                        cve_and_bug.extend(data[errata]['BUG'])
                        url = data[errata]['URL'] if data[errata]['URL'] else 'N/A'
                        tmp_content.append(
                            [package.name, errata, url, translate(severity), ' '.join(cve_and_bug), compatibility_abi,
                             compatibility_drive_kabi, is_compatible, undo, package.installed_pkg, package.install_pkg])
                if not package.erratas:
                    tmp_content.append(
                        [package.name, "N/A", "N/A", "N/A", "N/A", compatibility_abi, compatibility_drive_kabi, is_compatible,
                         undo, package.installed_pkg, package.install_pkg])

                # 处理compatibility_details
                index = 0
                if len(package.compatibility_details) > len(tmp_content):
                    for change, detail in package.compatibility_details.items():
                        installed, install, detail_msg = self._get_compatibility_details(change, detail)
                        if index > len(tmp_content) - 1:
                            tmp_content.append(
                                [package.name, "N/A", "N/A", "N/A", "N/A", compatibility_abi, compatibility_drive_kabi,
                                 is_compatible, undo, package.installed_pkg, package.install_pkg, installed, install,
                                 detail_msg])
                        else:
                            tmp_content[index].extend([installed, install, detail_msg])
                        index += 1
                else:
                    for change, detail in package.compatibility_details.items():
                        installed, install, detail_msg = self._get_compatibility_details(change, detail)
                        tmp_content[index].extend([installed, install, detail_msg])
                        index += 1
                # 对没有内容的字段填充N/A
                for line in tmp_content:
                    if len(line) == 11:
                        line.extend(['N/A', 'N/A', 'N/A'])
                write_content.extend(tmp_content)
            csvwriter.writerows(write_content)

        if not all([self.is_compatible, self.undo]):
            # 表示兼容性检查未通过，打印提示信息，并提示详细文件保存的路径
            print_str = translate("Compatibility check failed, please refer to the following details file: %s") % (
                self.details_file)
            logger.warning(print_str)
        else:
            print_str = translate("Compatibility check passed, no problem found. Details file: %s") % (
                self.details_file)
            logger.info(print_str)

    def _get_user_comfirm(self):
        """
        当存在兼容性问题时，获取用户的确认结果，如果执行时有 -y 参数，则默认为yes
        """
        if not all([self.is_compatible, self.undo]) and not self.conf.assumeyes:
            input_str = translate("Do you want to continue? [y/N]")
            output = Output(None, self.conf)
            confirm = output.userconfirm(msg=input_str, defaultyes_msg=input_str)
            if not confirm or self.conf.assumeno:
                raise CliError(translate("Operation aborted."))

    def config(self):
        """
        缓存的时候增加对自定义文件的下载支持
        """
        self.conf = self.base.conf
        all_repos = self.base.repos.all()
        for repo in all_repos:
            repo.add_metadata_type_to_download('compatibility')

    def pre_transaction(self):
        """
        在事务执行前进行兼容性检查
        """
        logger.info(translate("Running compatibility checks"))
        try:
            chang_packages_info = utils.get_change_packages_info(self.base)
            if chang_packages_info:
                # 比较每一个包的两个版本，判断是否是回退动作
                downgraged = False
                for name, value in chang_packages_info.items():
                    if downgraged:
                        break
                    if utils.compare_rpm_evr(value['old'].evr, value['new'].evr):
                        downgraged = True
                if downgraged:
                    apkg_adv_insts = utils.get_apkg_adv_insts(self.base, [value['new'] for _, value in chang_packages_info.items()])
                else:
                    apkg_adv_insts = utils.get_apkg_adv_insts(self.base)
                cannot_undo_packages = utils.get_cannot_undo_packages(self.base,
                                                                      [chang_packages_info[item]['old'] for item in
                                                                       chang_packages_info])
                abiinfo = utils.get_abiinfo_content(self.base)

                utils.generate_package_info(chang_packages_info, apkg_adv_insts, [x.name for x in cannot_undo_packages],
                                            abiinfo, self.package_info, downgraged)

                self._get_output(self.package_info)
                # 获取几个变量的值并赋值
                self.undo = utils.UNDO
                self.is_compatible = utils.IS_COMPATIBLE
                self.is_undo_flag = utils.IS_UNDO_FLAG

                self._write_check_result()
                self._get_user_comfirm()
        except CliError:
            raise CliError(translate("Operation aborted."))
        except Exception as e:
            logger.critical(
                translate("Some errors occurred while performing compatibility checks, error: %s") % (str(e)))
            raise CliError(translate("Operation aborted."))

        logger.info(translate("Compatibility checks completed."))
