import dnf
import rpm
import hawkey
import json

from utils.package_info import PackageInfo

# 定义几个状态变量
UNDO = True
IS_COMPATIBLE = True
IS_UNDO_FLAG = False


def get_change_packages_info(dnf_base):
    """
    获取本次升级或降级的两个版本的软件包信息
    :param dnf_base: dnf.Base对象
    """
    # 获取当前已安装的软件包信息
    installed_query = dnf_base.sack.query().installed()

    # 获取本次操作变更的软件包信息
    if hasattr(dnf.util, "_make_lists"):
        list_bunch = dnf.util._make_lists(dnf_base.transaction)
    else:
        # 处理3.4-5上dnf.util没有_make_lists的问题
        from dnf.cli import output
        list_bunch = output._make_lists(dnf_base.transaction, dnf_base._goal)
    change_packages_info = {}

    for _, pkg_list in list_bunch.items():
        for item in pkg_list:
            pkg_name = item.name
            if pkg_name == 'kernel' and _ == 'erased':  # 处理kernel包执行undo的情况
                old_pkg = installed_query.filter(name=pkg_name, epoch=item.epoch, version=item.version,
                                                 release=item.release, arch=item.arch).run()
                item = running_kernel_pkgs(dnf_base)[0]
            else:
                old_pkg = installed_query.filter(name=pkg_name).run()
            if old_pkg:
                change_packages_info[pkg_name] = {
                    'old': old_pkg[0],
                    'new': item
                }

    return change_packages_info


def compare_rpm_evr(evr1, evr2):
    """
    比较两个包版本号的大小
    :param evr1: 软件包1的epoch:version-release
    :param evr2: 软件包2的epoch:version-release
    """
    if ":" in evr1:
        epoch1 = evr1.split(":")[0]
        evr1 = evr1.split(":")[1]
    else:
        epoch1 = None
    version1, release1 = evr1.split("-")

    if ":" in evr2:
        epoch2 = evr2.split(":")[0]
        evr2 = evr2.split(":")[1]
    else:
        epoch2 = None
    version2, release2 = evr2.split("-")

    result = rpm.labelCompare((epoch1, version1, release1), (epoch2, version2, release2))

    if result > 0:
        # 勘误中的版本跟升级的版本大，则不需展示勘误信息
        return True
    else:
        return False


def init_objects(packages_erratas, cannot_undo_packages, packages_abiinfo, package_info, pkgs):
    """
    初始化每个PackageInfo类的实例
    :param packages_erratas: 此次变化软件包的勘误信息
    :param cannot_undo_packages: 此次变化软件包中无法回退的软件包名列表
    :param packages_abiinfo: 此次变化软件包的abi、kabi信息
    :param package_info: 用于存放PackageInfo对象的列表
    :param pkgs: 变化的软件包的信息，包括变化前和变化后的包对象
    """

    def _get_info(_pkg, _packages_abiinfo, _cannot_undo_packages, _pkgs):
        global UNDO, IS_COMPATIBLE
        # 判断是否能够回退
        if _pkg in _cannot_undo_packages:
            UNDO = False

        # 判断是否是兼容的
        _compatibility = _packages_abiinfo[_pkg]['compatibility']
        _drive_kabi = _packages_abiinfo[_pkg]['drive_kabi']
        _is_compatible = _packages_abiinfo[_pkg]['is_compatible']
        _compatibility_details = _packages_abiinfo[_pkg]['compatibility_details']
        _undo = False if _pkg in _cannot_undo_packages else True
        old_pkg = _pkgs[_pkg]['old']
        _installed_pkg = "{name}-{version}-{release}.{arch}".format(name=old_pkg.name, version=old_pkg.version,
                                                                    release=old_pkg.release, arch=old_pkg.arch)
        new_pkg = _pkgs[_pkg]['new']
        _install_pkg = "{name}-{version}-{release}.{arch}".format(name=new_pkg.name, version=new_pkg.version,
                                                                  release=new_pkg.release, arch=new_pkg.arch)
        if not _is_compatible:
            IS_COMPATIBLE = False

        return _compatibility, _drive_kabi, _is_compatible, _compatibility_details, _undo, _installed_pkg, _install_pkg

    for pkg, errata in packages_erratas.items():
        compatibility, drive_kabi, is_compatible, compatibility_details, undo, installed_pkg, install_pkg = _get_info(
            pkg, packages_abiinfo, cannot_undo_packages, pkgs)

        package_info.append(
            PackageInfo(pkg, errata, compatibility, drive_kabi, is_compatible, compatibility_details, undo,
                        installed_pkg, install_pkg))

    for pkg in packages_abiinfo:
        if pkg not in packages_erratas:
            compatibility, drive_kabi, is_compatible, compatibility_details, undo, installed_pkg, install_pkg = _get_info(
                pkg, packages_abiinfo, cannot_undo_packages, pkgs)

            package_info.append(
                PackageInfo(pkg, {}, compatibility, drive_kabi, is_compatible, compatibility_details, undo,
                            installed_pkg, install_pkg))


def get_change_package_abi_info(old_pkg, new_pkg, abiinfo):
    """
    获取指定包的两个版本的兼容性信息
    :param old_pkg: 当前系统上安装的包对象
    :param new_pkg: 将要安装的包对象
    :param abiinfo: 软件源中所有的compatibility信息
    """
    global IS_UNDO_FLAG
    package_abiinfo = {
        'compatibility': 1.00,
        'is_compatible': True,
        'compatibility_details': {},
    }
    srpm_name = old_pkg.source_name
    old_pkg_nvr_src = "{name}-{version}-{release}".format(name=srpm_name, version=old_pkg.version,
                                                          release=old_pkg.release)
    new_pkg_nvr_src = "{name}-{version}-{release}".format(name=srpm_name, version=new_pkg.version,
                                                          release=new_pkg.release)
    old_pkg_nvr_bin = "{name}-{version}-{release}".format(name=old_pkg.name, version=old_pkg.version,
                                                          release=old_pkg.release)
    new_pkg_nvr_bin = "{name}-{version}-{release}".format(name=new_pkg.name, version=new_pkg.version,
                                                          release=new_pkg.release)
    for item in abiinfo:
        if srpm_name in item:
            find_key = None
            find_key_update = "{old}:{new}".format(old=old_pkg_nvr_src, new=new_pkg_nvr_src)
            find_key_undo = "{new}:{old}".format(old=old_pkg_nvr_src, new=new_pkg_nvr_src)
            if find_key_update in item[srpm_name]:
                find_key = find_key_update
            elif find_key_undo in item[srpm_name]:
                find_key = find_key_undo
                IS_UNDO_FLAG = True

            if find_key:
                is_compatible = True if item[srpm_name][find_key]['check_result'] == "PASS" else False
                all_compatibility_details = item[srpm_name][find_key]['detail_results']
                if old_pkg.name == "kernel":
                    compatibility_details = all_compatibility_details
                else:
                    if old_pkg_nvr_bin in all_compatibility_details:
                        compatibility_details = all_compatibility_details[old_pkg_nvr_bin]
                    elif new_pkg_nvr_bin in all_compatibility_details:
                        compatibility_details = all_compatibility_details[new_pkg_nvr_bin]
                    else:
                        compatibility_details = {}

                if srpm_name == "kernel":
                    compatibility = item[srpm_name][find_key]['kabi']
                    compatibility_drive_kabi = item[srpm_name][find_key]['drive_kabi']
                else:
                    compatibility = item[srpm_name][find_key]['abi']
                    compatibility_drive_kabi = None

                package_abiinfo = {
                    'compatibility': compatibility,
                    'is_compatible': is_compatible,
                    'compatibility_details': compatibility_details,
                    'drive_kabi': compatibility_drive_kabi
                }

                break
    return package_abiinfo


def generate_package_info(pkgs, adv_insts, cannot_undo_packages, abiinfo, package_info, is_downgraged):
    """
    生成PackageInfo对象
    :param pkgs: 变化的软件包的信息，包括变化前和变化后的包对象
    :param adv_insts: 当前系统上已安装的包的勘误信息
    :param cannot_undo_packages: 此次变化中无法回退的软件包列表
    :param abiinfo: 软件源中所有的compatibility信息
    :param package_info: 用于存放PackageInfo对象的列表
    """
    packages_erratas = {}
    packages_abiinfo = {}
    for apkg, advisory in adv_insts:
        if apkg.name in pkgs:
            if is_downgraged:
                if not compare_rpm_evr(apkg.evr, pkgs[apkg.name]['new'].evr):
                    continue
            else:
                if compare_rpm_evr(apkg.evr, pkgs[apkg.name]['new'].evr):
                    continue
            severity = advisory.severity
            errata_id = advisory.id
            description = advisory.description
            references = advisory.references
            cve_info, bug_info = set(), set()
            errata_url = ""
            for ref in references:
                if ref.type == hawkey.REFERENCE_CVE:
                    cve_info.add(ref.id)
                elif ref.type == hawkey.REFERENCE_BUGZILLA:
                    bug_info.add(ref.id)
                elif ref.type == hawkey.REFERENCE_UNKNOWN:
                    errata_url = ref.url
            tmp = {
                errata_id: {
                    "CVE": list(cve_info),
                    "BUG": list(bug_info),
                    "URL": errata_url,
                    "description": description
                }
            }

            if apkg.name not in packages_erratas:
                packages_erratas[apkg.name] = {}
                packages_erratas[apkg.name][severity] = [tmp]
            else:
                if severity not in packages_erratas[apkg.name]:
                    packages_erratas[apkg.name][severity] = [tmp]
                else:
                    if errata_id not in [list(item.keys())[0] for item in packages_erratas[apkg.name][severity]]:
                        packages_erratas[apkg.name][severity].append(tmp)

    for pkg_name in pkgs:
        # 获取abi信息
        package_abiinfo = get_change_package_abi_info(pkgs[pkg_name]['old'], pkgs[pkg_name]['new'], abiinfo)
        if pkg_name not in packages_abiinfo:
            packages_abiinfo[pkg_name] = {
                'compatibility': package_abiinfo['compatibility'],
                'is_compatible': package_abiinfo['is_compatible'],
                'compatibility_details': package_abiinfo['compatibility_details']
            }
            if 'drive_kabi' in package_abiinfo:
                packages_abiinfo[pkg_name]['drive_kabi'] = package_abiinfo['drive_kabi']
            else:
                packages_abiinfo[pkg_name]['drive_kabi'] = None

    init_objects(packages_erratas, cannot_undo_packages, packages_abiinfo, package_info, pkgs)


def running_kernel_pkgs(dnf_base):
    """
    获取当前运行内核的包对象
    :param dnf_base: dnf.Base对象
    """
    sack = dnf_base.sack
    q = sack.query().filterm(empty=True)
    kernel = sack.get_running_kernel()
    if kernel:
        q = q.union(sack.query().filterm(sourcerpm=kernel.sourcerpm))
    return q


def get_apkg_adv_insts(dnf_base, undo_pkgs=None):
    """
    获取当前系统安装的所有包，并返回包和对应的勘误
    :param dnf_base: dnf.Base对象
    :param undo_pkgs: 执行undo时，指定将安装的软件包对象
    """
    if undo_pkgs:
        sack = dnf_base.sack
        q = sack.query().filterm(empty=True)
        # 将当前运行内核的查询对象增加到查询对象中
        q = q.union(running_kernel_pkgs(dnf_base).installed())
        for pkg in undo_pkgs:
            q = q.union(dnf_base.sack.query().available().filter(name=pkg.name, version=pkg.version, release=pkg.release, arch=pkg.arch))
        for apackage in q.get_advisory_pkgs(hawkey.GT):
            advisory = apackage.get_advisory(dnf_base.sack)
            yield apackage, advisory

    else:
        q = dnf_base.sack.query().installed().latest(1)
        # 将当前运行内核的查询对象增加到查询对象中
        q = q.union(running_kernel_pkgs(dnf_base).installed())
        for apackage in q.get_advisory_pkgs(hawkey.GT):
            advisory = apackage.get_advisory(dnf_base.sack)
            yield apackage, advisory


def get_cannot_undo_packages(dnf_base, pkgs):
    """
    获取无法回滚的软件包列表
    :param dnf_base: dnf.Base对象
    :param pkgs: 此次变化中将要变化的包对象列表
    """
    available_query = dnf_base.sack.query().available()
    cannot_undo_packages = []
    for pkg in pkgs:
        q = available_query.filter(name=pkg.name, version=pkg.version, release=pkg.release, arch=pkg.arch).run()
        if not q:
            cannot_undo_packages.append(pkg)
    return cannot_undo_packages


def get_abiinfo_content(dnf_base):
    """
    获取软件源中的abiinfo信息
    :param dnf_base: dnf.Base对象
    """
    all_repos = dnf_base.repos.all()
    abiinfo_content = []
    for repo in all_repos:
        tmp_abiinfo_content = repo.get_metadata_content("compatibility")
        if tmp_abiinfo_content:
            abiinfo_content.append(json.loads(tmp_abiinfo_content))
    return abiinfo_content
